/**@name LIBRARY: unSystemIntegrity_LHCLoggingDB.ctl

@author:
Josef Hofer (EN-ICE-SCD)

Creation Date: 23/06/2014

Purpose:
 System integrity plugin for the unLHCLoggingDB component.

Usage: Public

PVSS manager usage: CTRL

*/

//--------------------------------------------------------------------------------------------------------------------------------
// constant declaration
//------------------

// String constants for adressing datapoints
const string c_SysInt_LHCLoggingDB = "LHCLoggingDB";
const string LHCLoggingDB_pattern = "LHCLoggingDB";
const string LDB_pattern = "LDB"; // shorter pattern for LHCLoggingDB

// Additional commands for function unSystemIntegrity_LHCLoggingDB_HandleCommand
const int c_SysInt_LHCLoggingDB_CreateTestDps = 20;
const int c_SysInt_LHCLoggingDB_RemoveTestDps = 21;

// Indexes of settings in .config.data
const int c_SysInt_LHCLoggingDB_ConfigIndex_Count                 = 5;
const int c_SysInt_LHCLoggingDB_ConfigIndex_LoopDelay             = 1;
const int c_SysInt_LHCLoggingDB_ConfigIndex_MaxAgeTransferJobs    = 2;
const int c_SysInt_LHCLoggingDB_ConfigIndex_MaxAgeTransferHistory = 3;
const int c_SysInt_LHCLoggingDB_ConfigIndex_RegistrationAttempts  = 4;
const int c_SysInt_LHCLoggingDB_ConfigIndex_RegistrationWaitTime  = 5;

// Default values for .config.data
const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_LoopDelay             = 240;  // [s]
const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferJobs    = 360;  // [s]
const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferHistory = 360;  // [s]
const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationAttempts  = 5;
const unsigned c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationWaitTime  = 3;    // [s]

// Default values for hard coded settings
// ...

// Message indexes in LHCLoggingDB catalogue
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorNotFound             = 1;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorLoadDbSettings       = 2;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorOpenDbConnection     = 3;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorDbSchemaNotFound     = 4;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps      = 5;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorDeletingTestDps      = 6;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorRegisteringTestDps   = 7;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorUnregisteringTestDps = 8;
const int c_SysInt_LHCLoggingDB_Catalogue_ConfigDataInconsistent    = 9;
const int c_SysInt_LHCLoggingDB_Catalogue_ConfigDataHasBeenReset    = 10;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorStoppingThread       = 11;
const int c_SysInt_LHCLoggingDB_Catalogue_ErrorFullConsistencyCheck = 12;

// Integrity classes
const string c_SysInt_LHCLoggingDB_IntegrityClass_TransferJobs     = "LHCLoggingDB_TransferJobs";
const string c_SysInt_LHCLoggingDB_IntegrityClass_TransferHistory  = "LHCLoggingDB_TransferHistory";
const string c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints   = "LHCLoggingDB_TestDatapoints";
const string c_SysInt_LHCLoggingDB_IntegrityClass_DataConsistency  = "LHCLoggingDB_DataConsistency";

// Alert levels of the Integrity classes
const unsigned c_SysInt_LHCLoggingDB_TransferJobs_ErrorQuery            = 70;
const unsigned c_SysInt_LHCLoggingDB_TransferJobs_ErrorLagging          = 20;
const unsigned c_SysInt_LHCLoggingDB_TransferJobs_ErrorDisabled         = 10;
const unsigned c_SysInt_LHCLoggingDB_TransferJobs_OK                    =  0;

const unsigned c_SysInt_LHCLoggingDB_TransferHistory_ErrorQuery         = 70;
const unsigned c_SysInt_LHCLoggingDB_TransferHistory_ErrorRegistration  = 30;
const unsigned c_SysInt_LHCLoggingDB_TransferHistory_ErrorTransfer      = 20;
const unsigned c_SysInt_LHCLoggingDB_TransferHistory_WarningLagging     = 10;
const unsigned c_SysInt_LHCLoggingDB_TransferHistory_OK                 =  0;

const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery          = 70;
const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration   = 40;
const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_ErrorTransfer       = 30;
const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_Retransfering       = 20;
const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_Initialising        = 10;
const unsigned c_SysInt_LHCLoggingDB_TestDatapoints_OK                  =  0;

const unsigned c_SysInt_LHCLoggingDB_DataConsistency_ErrorQuery         = 70;
const unsigned c_SysInt_LHCLoggingDB_DataConsistency_Inconsistent       = 10;
const unsigned c_SysInt_LHCLoggingDB_DataConsistency_OK                 =  0;

// String constants
const string c_SysInt_LHCLoggingDB_TestDatapoint_Description = "System integrity test datapoint for LHCLoggingDB";

//--------------------------------------------------------------------------------------------------------------------------------
// declaration of global variables
//------------------

global dyn_string g_SysInt_LHCLoggingDB_ActiveIntegrityClasses;   // only needed for panel to show activated integrity classes
global string g_SysInt_LHCLoggingDB_mySystemDbSchema;             // db schema of this system
global int g_SysInt_LHCLoggingDB_monitoringThreadId;              // ID of the monitoring thread

global unsigned g_SysInt_LHCLoggingDB_loopDelay;                  // [s] run checks according to this value
global unsigned g_SysInt_LHCLoggingDB_maxAgeTransferJobs;         // [s] transfer jobs have to run within this time
global unsigned g_SysInt_LHCLoggingDB_maxAgeTransferHistory;      // [s] transfer history has to be updated within this time
global unsigned g_SysInt_LHCLoggingDB_registrationAttempts;       // maximum attempts to read ELEMENTS table before registering
global unsigned g_SysInt_LHCLoggingDB_registrationWaitTime;       // wait time after unsucessfull attempt to read ELEMENTS table before doing next attempt

//@{

//---------------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_Initialize
/**
* Initialise system integrity plugin for LHCLoggingDB.
*
* @param dsResult   output, not used
*/
public void unSystemIntegrity_LHCLoggingDB_Initialize(dyn_string &dsResult)
{
    unSystemIntegrity_LHCLoggingDB_getConfigData(true);

    unSystemIntegrity_LHCLoggingDB_initDabatase();

    // Look for active integrity classes and write them into global list
    dyn_string dsSysIntAlarmDps = dpNames(c_unSystemAlarm_dpPattern + LHCLoggingDB_pattern + "*", c_unSystemAlarm_dpType);
    int iDsSysIntAlarmDpsLen = dynlen(dsSysIntAlarmDps);
    for (int i = 1; i <= iDsSysIntAlarmDpsLen; i++) {
        string sAlarmDpToCheck = dpSubStr(dsSysIntAlarmDps[i], DPSUB_DP);
        dynAppend(g_SysInt_LHCLoggingDB_ActiveIntegrityClasses, sAlarmDpToCheck);
    }
}



//---------------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_HandleCommand
/**
* Handle systemIntegrity commands from configuration panel
*
* @param sDpe1      input, dpe
* @param command    input, the systemIntegrity command
* @param sDpe2      input, dpe
* @param parameters input, the list of parameters of the systemIntegrity command
*/
public void unSystemIntegrity_LHCLoggingDB_HandleCommand(string sDpe1, int command, string sDpe2, dyn_string parameters)
{
    dyn_string dsExInfo;

    switch (command) {
        case UN_SYSTEMINTEGRITY_ADD: { // add the _UnSystemAlarm DP
            int iParametersLen = dynlen(parameters);
            for (int i = 1; i <= iParametersLen; i++) {
                unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(parameters[i], true, false, true, dsExInfo);

                if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", parameters[i])) {
                    unSystemIntegrity_LHCLoggingDB_createTestDPs();
                    unSystemIntegrity_LHCLoggingDB_registerTestDPs();
                }
            }
            // ensure that thread is running
            unSystemIntegrity_LHCLoggingDB_manageMonitoringThread("start");
            break;
        }
        case UN_SYSTEMINTEGRITY_DELETE: { // delete the _UnSystemAlarm DP
            int iParametersLen = dynlen(parameters);
            for (int i = 1; i <= iParametersLen; i++) {
                unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(parameters[i], false, true, false, dsExInfo);

                if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", parameters[i])) {
                    unSystemIntegrity_LHCLoggingDB_unregisterTestDPs();
                    unSystemIntegrity_LHCLoggingDB_deleteTestDPs();
                }
            }
            delay(1);
            // stop thread if necessary
            unSystemIntegrity_LHCLoggingDB_manageMonitoringThread("stop");
            break;
        }
        case UN_SYSTEMINTEGRITY_ENABLE: { // enable the _UnSystemAlarm DP
            int iParametersLen = dynlen(parameters);
            for (int i = 1; i <= iParametersLen; i++) {
                unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(parameters[i], false, false, true, dsExInfo);

                if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", parameters[i])) {
                    unSystemIntegrity_LHCLoggingDB_createTestDPs();
                    unSystemIntegrity_LHCLoggingDB_registerTestDPs();
                }
            }
            // ensure that thread is running
            unSystemIntegrity_LHCLoggingDB_manageMonitoringThread("start");
            break;
        }
        case UN_SYSTEMINTEGRITY_DISABLE: { // disable the _UnSystemAlarm DP
            int iParametersLen = dynlen(parameters);
            for (int i = 1; i <= iParametersLen; i++) {
                unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(parameters[i], false, false, false, dsExInfo);
            }
            delay(1);
            // stop thread if necessary
            unSystemIntegrity_LHCLoggingDB_manageMonitoringThread("stop");
            break;
        }
        case UN_SYSTEMINTEGRITY_DIAGNOSTIC: { // give the list of _UnSystemAlarm DP and the state
            dpSetWait(c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", command,
                      c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result", g_SysInt_LHCLoggingDB_ActiveIntegrityClasses);
            break;
        }
        case UN_SYSTEMINTEGRITY_EXTENDED_DIAGNOSTIC: {
            unSystemIntegrity_LHCLoggingDB_doFullConsistencyCheck(dsExInfo);
            string sAnswer = "ok";
            if (dynlen(dsExInfo) > 0) {
                sAnswer = "error";
            }
            dpSetWait(c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.commandResult", command,
                      c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".diagnostic.result", makeDynString(sAnswer));
            break;
        }
        case c_SysInt_LHCLoggingDB_CreateTestDps: {
            DebugTN("creation of test dps was triggered");

            if (dpExists(c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints)) {
                unSystemIntegrity_LHCLoggingDB_createTestDPs();
                unSystemIntegrity_LHCLoggingDB_registerTestDPs();
            }
            break;
        }
        case c_SysInt_LHCLoggingDB_RemoveTestDps: {
            DebugTN("destruction of test dps was triggered");

            unSystemIntegrity_LHCLoggingDB_unregisterTestDPs();
            unSystemIntegrity_LHCLoggingDB_deleteTestDPs();
            break;
        }
//  case UN_SYSTEMINTEGRITY_DEBUG:
//  case UN_SYSTEMINTEGRITY_VERSION:
        default: {
            break;
        }
    }

    if (dynlen(dsExInfo) > 0) {
        if (isFunctionDefined("unMessageText_sendException")) {
            unMessageText_sendException("*", "*", g_SystemIntegrity_sSystemName + "unSystemIntegrity_LHCLoggingDB_HandleCommand", "user", "*", dsExInfo);
        }
        // handle any error in case the send message failed
        if (dynlen(dsExInfo) > 0) {
            DebugTN(g_SystemIntegrity_sSystemName + "unSystemIntegrity_LHCLoggingDB_HandleCommand", dsExInfo);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_DataCallback
/**
* Callback function to get most recent configuration data
*
* @param sDPe
* @param dsResult
*/
public void unSystemIntegrity_LHCLoggingDB_DataCallback(string sDPe, dyn_string &dsResult)
{
    unSystemIntegrity_LHCLoggingDB_getConfigData(false);
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_postInstall
/**
* Create systemIntegrity datapoint for LHCLoggingDB
*/
public void unSystemIntegrity_LHCLoggingDB_postInstall()
{
    DebugTN("unSystemIntegrity_LHCLoggingDB_postInstall(): performing post installation steps");

    // Create system integrity datapoint for LHC Logging DB
    const string sSysIntDp = c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION;
    if (!dpExists(sSysIntDp)) {
        dpCreate(sSysIntDp, UN_SYSTEMINTEGRITY_DPTYPE);

        dpSetWait(
            sSysIntDp + ".config.functions",
            makeDynString("unSystemIntegrity_LHCLoggingDB_Initialize"
                          , "unSystemIntegrity_LHCLoggingDB_HandleCommand"
                          , "unSystemIntegrity_LHCLoggingDB_DataCallback"),
            sSysIntDp + ".config.configurationPanel",
            "vision/systemIntegrity/systemIntegrity_LHCLoggingDB_configuration.pnl",
            sSysIntDp + ".config.operationPanel",
            "vision/systemIntegrity/systemIntegrity_LHCLoggingDB_operation.pnl"
        );
    }

    // Get settings and reset them if they are not correct
    unSystemIntegrity_LHCLoggingDB_getConfigData(true);

    // Recreate alert config of LHC Logging Db system alarm dps to update it in case of change in structure
    dyn_string dsAlarmDps = dpNames(c_unSystemAlarm_dpPattern + LHCLoggingDB_pattern + "*", c_unSystemAlarm_dpType);
    dyn_errClass err = getLastError();
    if (dynlen(err) == 0) {
        int iDsAlarmDpsLen = dynlen(dsAlarmDps);
        for (int i = 1; i <= iDsAlarmDpsLen; i++) {
            dyn_string dsExInfo;
            fwAlertConfig_delete(dsAlarmDps[i] + ".alarm", dsExInfo) ;
            unSystemIntegrity_LHCLoggingDB_addAlertConfig(dsAlarmDps[i] + ".alarm", dsExInfo);
        }
    }
}


//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_initDabatase
/**
* Get database settings, db schema and check if db schema exists in database
*/
private void unSystemIntegrity_LHCLoggingDB_initDabatase()
{
    string sDbName, sDbUser, sDbPass;

    // Get database settings
    if (unLHCLoggingDBSettings_getActiveDbConfig(sDbName, sDbUser, sDbPass) == false) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorLoadDbSettings);
        return;
    }

    // Connect to database
    if (unLHCLoggingDBSql_openDbConnection(sDbName, sDbUser, sDbPass) == false) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorOpenDbConnection);
        return;
    }

    // Get db schemas of the connected db
    dyn_string dsExInfo;
    dyn_string dsDbSchemas = unLHCLoggingDBSql_getDbSchemas(dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorDbSchemaNotFound);
        return;
    }

    // Get db schema of this system
    g_SysInt_LHCLoggingDB_mySystemDbSchema = unLHCLoggingDBSettings_getMySystemDbSchema();

    // Check if schema is known by the database
    if (dynContains(dsDbSchemas, g_SysInt_LHCLoggingDB_mySystemDbSchema) <= 0) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorDbSchemaNotFound);
        return;
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_createTestDPs
/**
* Create a test datapoint for each data category, if it does not exist yet
*/
private void unSystemIntegrity_LHCLoggingDB_createTestDPs()
{
    // Get data categories for which we will create test datapoints
    dyn_string dsExInfo;
    dyn_string dsDataCategories = unLHCLoggingDBSql_getDataCategoriesOfSchema(g_SysInt_LHCLoggingDB_mySystemDbSchema, dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError(
                "ERROR", 
                c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps, 
                "Data categories could not be read"
        );
        return;
    }

    // Create test datapoints
    int len = dynlen(dsDataCategories);
    for (int i = 1; i <= len; i++) {
        string sTestDP = c_SysInt_RDB_TestDp_Prefix + "_" + LDB_pattern + "_" + dsDataCategories[i];
        string sTestDPold = c_SysInt_RDB_TestDp_Prefix + "_" + LHCLoggingDB_pattern + "_" + dsDataCategories[i];
               
        if (!dpExists(sTestDP) && !dpExists(sTestDPold)) { // only create new test dp, if there is no existing one         
            dpCreate(sTestDP, c_SysInt_RDB_TestDp_Type);
            if (!dpExists(sTestDP)) {
                unSystemIntegrity_LHCLoggingDB_throwError(
                        "ERROR", 
                        c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps, 
                        "Datapoint " + sTestDP + " not created"
                );
                return;
            }
            dpSetDescription(sTestDP + c_SysInt_RDB_TestDpe_Value, c_SysInt_LHCLoggingDB_TestDatapoint_Description);            
            unSystemIntegrity_LHCLoggingDB_addArchiveConfig(sTestDP + c_SysInt_RDB_TestDpe_Value);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_registerTestDPs
/**
* Register test dapaoints to logging
*/
private void unSystemIntegrity_LHCLoggingDB_registerTestDPs()
{
    dyn_string dsDpe, dsAlias, dsVarName, dsDescription, dsDataCategoryId, dsHierarchy;
    dyn_string dsElementId, dsUnit, dsRDBSource, dsErrors, dsWarnings;
    dyn_string dsExInfo;

    dyn_string dsTestDp = unSystemIntegrity_LHCLoggingDB_getTestDps();  

    int len = dynlen(dsTestDp);
    for (int i = 1; i <= len; i++) {
        dynAppend(dsVarName,  unSystemIntegrity_LHCLoggingDB_getVarName(dsTestDp[i]));
        dynAppend(dsDpe, dpSubStr(dsTestDp[i] + c_SysInt_RDB_TestDpe_Value, DPSUB_SYS_DP_EL));
        dynAppend(dsDescription, c_SysInt_LHCLoggingDB_TestDatapoint_Description);
        dynAppend(dsHierarchy, "");
        dynAppend(dsAlias, "");
        string sDataCategory = unSystemIntegrity_LHCLoggingDB_getDataCategory(dsTestDp[i]);
        dynAppend(dsDataCategoryId, unLHCLoggingDBSql_getDataCategoryId(sDataCategory, dsExInfo));
    }

    if (dynlen(dsDpe) > 0) {
        unsigned attempts = 0;
        while (true) {
            if (unLHCLoggingDBParser_getElementData(dsDpe, dsDpe, dsAlias, dsElementId, dsUnit, dsRDBSource, dsErrors, dsWarnings)) { break; }
            if (++attempts > g_SysInt_LHCLoggingDB_registrationAttempts) {
                unSystemIntegrity_LHCLoggingDB_throwError(
                        "ERROR", 
                        c_SysInt_LHCLoggingDB_Catalogue_ErrorRegisteringTestDps, 
                        "Error retrieving ELEMENT_ID"
                );
                return;
            }
            delay(g_SysInt_LHCLoggingDB_registrationWaitTime);
        }

        string sUser, sPass;
        unLHCLoggingDBSql_getSystemIntegrityCredentials(sUser, sPass, dsExInfo);
        if (dynlen(dsExInfo) > 0) {
            unSystemIntegrity_LHCLoggingDB_throwError(
                    "ERROR", 
                    c_SysInt_LHCLoggingDB_Catalogue_ErrorRegisteringTestDps, 
                    "Error getting credentials"
            );
            return;
        }

        if (false == unLHCLoggingDBSql_registerDataForLogging(dsElementId, dsVarName, dsRDBSource, dsUnit, dsDescription, dsDataCategoryId, dsHierarchy, sUser, sPass)) {
            unSystemIntegrity_LHCLoggingDB_throwError(
                    "ERROR", 
                    c_SysInt_LHCLoggingDB_Catalogue_ErrorRegisteringTestDps, 
                    "Error during registration"
            );
            return;
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_deleteTestDPs
/**
* Delete test datapoints for each data category
*/
private void unSystemIntegrity_LHCLoggingDB_deleteTestDPs()
{
    dyn_string dsTestDP = unSystemIntegrity_LHCLoggingDB_getTestDps();  

    int len = dynlen(dsTestDP);
    for (int i = 1; i <= len; i++) {
        dpDelete(dsTestDP[i]);
        dyn_errClass err = getLastError();
        if (dynlen(err) > 0) {
            unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorDeletingTestDps);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_unregisterTestDPs
/**
* Unregister test dapaoints from logging
*/
private void unSystemIntegrity_LHCLoggingDB_unregisterTestDPs()
{
    dyn_string dsVariableName = unSystemIntegrity_LHCLoggingDB_getTestDpVarNames();    

    int len = dynlen(dsVariableName);
    for (int i = 1; i <= len; i++) {
        if (unLHCLoggingDBSql_unregisterVariableFromLogging(dsVariableName[i]) == false) {
            unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorUnregisteringTestDps);
            return;
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

/**
Add archive config for archive group EVENT to the test dpe 

@param sTestDpe, input, name of the test datapoint 
*/
private void unSystemIntegrity_LHCLoggingDB_addArchiveConfig(string sTestDpe) {       
    // get archive class for "EVENT"
    string sArchiveClass = unSystemIntegrity_LHCLoggingDB_getArchiveClass("EVENT");
    if (sArchiveClass == "") { return; } // no archive class found     
    
    // check if there is already an archive config
    int iArchiveType = 0;
    dpGet(sTestDpe + ":_archive.._type", iArchiveType);
    if (iArchiveType != 0) { return; } // archive config already exists 
    
    // create archive config
    dyn_string dsExInfo;
    fwArchive_set(sTestDpe, sArchiveClass, DPATTR_ARCH_PROC_VALARCH, 0, 0, 0, dsExInfo);
    if (dynlen(dsExInfo)) {
        unSystemIntegrity_LHCLoggingDB_throwError(
                "ERROR", 
                c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps, 
                "Archive config not created for " + sTestDpe
        );
    }  
}

/**
Get the name of the archive class which matches the pattern 

@param sPattern, input, pattern of the archive class like 'EVENT'  
@return name of the archive class
*/
private string unSystemIntegrity_LHCLoggingDB_getArchiveClass(string sPattern) {
    
    // Get all archive classes
    dyn_string dsArchiveClasses, dsArchiveGroupDps, dsExInfo;
    fwArchive_getAllRDBArchiveClasses(makeDynString(getSystemName()), dsArchiveClasses, dsArchiveGroupDps, dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError(
                "ERROR", 
                c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps, 
                "Archive groups could not be read"
        );
        return "";
    }

    // Check if there is an archive class that matches the pattern
    string sArchiveClass = "";
    int iDsArchiveClassesLen = dynlen(dsArchiveClasses);
    for (int i = 1; i <= iDsArchiveClassesLen; i++) {
        if (patternMatch("*"+sPattern+"*", dsArchiveClasses[i])) {
            sArchiveClass = dsArchiveClasses[i];
            break;
        }
    }
    if (sArchiveClass == "") {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorCreatingTestDps, "Archive group 'EVENT' not found");
        return "";
    }

    return sArchiveClass;    
}

/**
Get list of all variable names
  
@return list of test datapoints
*/
private dyn_string unSystemIntegrity_LHCLoggingDB_getTestDpVarNames() { 
    dyn_string dsVariableName;    
    dyn_string dsTestDp = unSystemIntegrity_LHCLoggingDB_getTestDps();
    
    // Calculate the variable name and append it to the result list
    int len = dynlen(dsTestDp);
    for (int i = 1; i <= len; i++) {
      string sVarName = unSystemIntegrity_LHCLoggingDB_getVarName(dsTestDp[i]);
      dynAppend(dsVariableName, sVarName);
    }
    
    return dsVariableName;  
}

/**
Get list of all test datapoints 
  
@return dsTestDp, list of test datapoints
*/
private dyn_string unSystemIntegrity_LHCLoggingDB_getTestDps() {   
   dyn_string dsTestDp = dpNames(c_SysInt_RDB_TestDp_Prefix + "_" + LHCLoggingDB_pattern + "_" + "*", c_SysInt_RDB_TestDp_Type);  
   dyn_string dsTestDp2= dpNames(c_SysInt_RDB_TestDp_Prefix + "_" + LDB_pattern + "_" + "*", c_SysInt_RDB_TestDp_Type);   
   dynAppend(dsTestDp, dsTestDp2);    
   return dsTestDp;
}

/**
Calculate the variable name from the dpe name of the test datapoint 

@param sDpeName, input, name of the test dp    
@return variable name 
*/
private string unSystemIntegrity_LHCLoggingDB_getVarName(string sTestDpName) {    
    // Remove system name
    sTestDpName = dpSubStr(sTestDpName,DPSUB_DP); 
    dyn_string dsSplit = strsplit(sTestDpName,"_");
    
    // Extract 'LHCLoggingDB' or 'LDB' pattern
    string sPattern = dsSplit[2]; 
 
    // Extract the data category
    string sDataCategory = unSystemIntegrity_LHCLoggingDB_getDataCategory(sTestDpName);     
    strreplace(sDataCategory, "/", "_"); // replace '/' by '_' because we don't use '/' for separation in DB     
    
    // Return the variable name
    return ( c_SysInt_RDB_TestDp_Prefix + "." + sPattern + "." + getSystemName() + sDataCategory );    
}

/**
Extract data category string from test datapoint name

@param sDpeName, input, name of the test dp    
@return data category 
*/
private string unSystemIntegrity_LHCLoggingDB_getDataCategory(string sTestDpName) {    
    // Remove system name
    sTestDpName = dpSubStr(sTestDpName,DPSUB_DP); 
    dyn_string dsSplit = strsplit(sTestDpName,"_");
    
    // Extract 'LHCLoggingDB' or 'LDB' pattern
    string sPattern = dsSplit[2]; 
 
    // Return the data category
    return (substr(sTestDpName, strlen(c_SysInt_RDB_TestDp_Prefix + "_" + sPattern + "_")));   
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_getConfigData
/**
* Get configuration data and settings from RDB_systemIntegrityInfo
*
* @param   bResetOnError   input,  true: set default values on error, false: don't set default values on error
*/
private void unSystemIntegrity_LHCLoggingDB_getConfigData(bool bResetOnError)
{
    if (dpExists(c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION)) {
        dyn_string dsConfigData;
        dpGet(c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".config.data", dsConfigData);

        if (dynlen(dsConfigData) == c_SysInt_LHCLoggingDB_ConfigIndex_Count) {
            g_SysInt_LHCLoggingDB_loopDelay             = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_LoopDelay];
            g_SysInt_LHCLoggingDB_maxAgeTransferJobs    = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_MaxAgeTransferJobs];
            g_SysInt_LHCLoggingDB_maxAgeTransferHistory = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_MaxAgeTransferHistory];
            g_SysInt_LHCLoggingDB_registrationAttempts  = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_RegistrationAttempts];
            g_SysInt_LHCLoggingDB_registrationWaitTime  = dsConfigData[c_SysInt_LHCLoggingDB_ConfigIndex_RegistrationWaitTime];
        } else {
            if (bResetOnError) {
                dpSetWait(c_SysInt_LHCLoggingDB + UN_SYSTEMINTEGRITY_EXTENSION + ".config.data",
                          makeDynString(c_SysInt_LHCLoggingDB_DefaultConfig_LoopDelay
                                        , c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferJobs
                                        , c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferHistory
                                        , c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationAttempts
                                        , c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationWaitTime)
                         );
                g_SysInt_LHCLoggingDB_loopDelay             = c_SysInt_LHCLoggingDB_DefaultConfig_LoopDelay;
                g_SysInt_LHCLoggingDB_maxAgeTransferJobs    = c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferJobs;
                g_SysInt_LHCLoggingDB_maxAgeTransferHistory = c_SysInt_LHCLoggingDB_DefaultConfig_MaxAgeTransferHistory;
                g_SysInt_LHCLoggingDB_registrationAttempts  = c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationAttempts;
                g_SysInt_LHCLoggingDB_registrationWaitTime  = c_SysInt_LHCLoggingDB_DefaultConfig_RegistrationWaitTime;

                unSystemIntegrity_LHCLoggingDB_throwError("INFO", c_SysInt_LHCLoggingDB_Catalogue_ConfigDataHasBeenReset);
            } else {
                unSystemIntegrity_LHCLoggingDB_throwError("WARNING", c_SysInt_LHCLoggingDB_Catalogue_ConfigDataInconsistent);
            }
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_manageIntegrityClass
/**
* This function is used to create/destruct/enable/disable alarm datapoints
* It can also create/destruct the test datapoints
*
* @param sDp             input, name of the alert dp
* @param bCreate         input, create dp (true: yes; false: no)
* @param bDelete         input, delete dp (true: yes; false: no)
*	@param bEnable         input, true: enable alarm; false: disable alarm
*	@param dsExInfo   output, exception are returned here
*/
private void unSystemIntegrity_LHCLoggingDB_manageIntegrityClass(string sDp, bool bCreate, bool bDelete, bool bEnable, dyn_string &dsExInfo)
{
    // Remove the system name
    string sSysAlarmDp = unSystemIntegrity_removeSystemName(sDp);

    if (bCreate) {
        unSystemIntegrity_LHCLoggingDB_createAlertDp(sSysAlarmDp);
        unSystemIntegrity_LHCLoggingDB_addAlertConfig(sSysAlarmDp + ".alarm", dsExInfo);

        if (dynlen(dsExInfo) <= 0) {
            // add to active integrity classes
            _unSystemIntegrity_setList(g_SysInt_LHCLoggingDB_ActiveIntegrityClasses, sSysAlarmDp, true);
        }
    }

    // Do enabling and masking
    if (dynlen(dsExInfo) <= 0) {
        if (dpExists(sSysAlarmDp)) {
            dpSetWait(sSysAlarmDp + ".enabled", bEnable);
            unAlarmConfig_mask(sSysAlarmDp + ".alarm", !bEnable, dsExInfo);
        }
    }

    if (bDelete) {
        // remove dp from the alert list of applicationDP
        _unSystemIntegrity_modifyApplicationAlertList(sSysAlarmDp, false, dsExInfo);

        // remove alert class from list
        _unSystemIntegrity_setList(g_SysInt_LHCLoggingDB_ActiveIntegrityClasses, sSysAlarmDp, false);

        // delete alarm datapoint
        if (dpExists(sSysAlarmDp)) {
            dpDelete(sSysAlarmDp);
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_createAlertDp
/**
* Create system integrity alert datapoint for LHC Logging DB
*
* @param sSysAlarmDp    input, alarm dp to create
*/
private void unSystemIntegrity_LHCLoggingDB_createAlertDp(string sSysAlarmDp)
{
    if (!dpExists(sSysAlarmDp)) {
        dpCreate(sSysAlarmDp, c_unSystemAlarm_dpType);

        string sSysAlarmDpe = sSysAlarmDp + ".alarm";
        if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TransferJobs + "*", sSysAlarmDpe)) {
            dpSetDescription(sSysAlarmDpe, "LHCLoggingDB transfer jobs:");
        } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TransferHistory + "*", sSysAlarmDpe)) {
            dpSetDescription(sSysAlarmDpe, "LHCLoggingDB transfer history:");
        } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", sSysAlarmDpe)) {
            dpSetDescription(sSysAlarmDpe, "LHCLoggingDB test datapoints:");
        } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_DataConsistency + "*", sSysAlarmDpe)) {
            dpSetDescription(sSysAlarmDpe, "LHCLoggingDB data consistency:");
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_RDB_addAlertConfig
/**
* Add discrete alert config to system alarm datapoint element
*
* @param sSysAlarmDpe    input, dpe for which we want to create the alert config
*	@param dsExInfo   output, exception are returned here
*/
private void unSystemIntegrity_LHCLoggingDB_addAlertConfig(string sSysAlarmDpe, dyn_string &dsExInfo)
{
    dyn_string dsAlertTexts;
    dyn_string dsAlertMatch;
    dyn_string dsAlertClasses;
    dyn_bool dbDiscreteNegation;
    dyn_string dsStateMatch;

    if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TransferJobs + "*", sSysAlarmDpe)) {
        dsAlertTexts = makeDynString(
                           "OK",
                           "Disabled transfer jobs found",
                           "Transfer job not executed in time",
                           "Query Error");
        dsAlertMatch = makeDynString(
                           "*",
                           c_SysInt_LHCLoggingDB_TransferJobs_ErrorDisabled,
                           c_SysInt_LHCLoggingDB_TransferJobs_ErrorLagging,
                           c_SysInt_LHCLoggingDB_TransferJobs_ErrorQuery);
        dsAlertClasses = makeDynString(
                             "",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.");
        dbDiscreteNegation = makeDynBool(false, false, false, false);
        dsStateMatch = makeDynString("", "", "", "");
    } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TransferHistory + "*", sSysAlarmDpe)) {
        dsAlertTexts = makeDynString(
                           "OK",
                           "Transfer lagging",
                           "Transfer Error",
                           "Registration Error",
                           "Query Error");
        dsAlertMatch = makeDynString(
                           "*",
                           c_SysInt_LHCLoggingDB_TransferHistory_WarningLagging,
                           c_SysInt_LHCLoggingDB_TransferHistory_ErrorTransfer,
                           c_SysInt_LHCLoggingDB_TransferHistory_ErrorRegistration,
                           c_SysInt_LHCLoggingDB_TransferHistory_ErrorQuery);
        dsAlertClasses = makeDynString(
                             "",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl1.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.");
        dbDiscreteNegation = makeDynBool(false, false, false, false, false);
        dsStateMatch = makeDynString("", "", "", "", "");
    } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints + "*", sSysAlarmDpe)) {
        dsAlertTexts = makeDynString(
                           "OK",
                           "Initialising",
                           "Retransfer started",
                           "Transfer Error",
                           "Registration Error",
                           "Query Error");
        dsAlertMatch = makeDynString(
                           "*",
                           c_SysInt_LHCLoggingDB_TestDatapoints_Initialising,
                           c_SysInt_LHCLoggingDB_TestDatapoints_Retransfering,
                           c_SysInt_LHCLoggingDB_TestDatapoints_ErrorTransfer,
                           c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration,
                           c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery);
        dsAlertClasses = makeDynString(
                             "",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl1.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl1.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.");
        dbDiscreteNegation = makeDynBool(false, false, false, false, false, false);
        dsStateMatch = makeDynString("", "", "", "", "", "");
    } else if (patternMatch("*" + c_SysInt_LHCLoggingDB_IntegrityClass_DataConsistency + "*", sSysAlarmDpe)) {
        dsAlertTexts = makeDynString(
                           "OK",
                           "Inconsistencies found",
                           "Query Error");
        dsAlertMatch = makeDynString(
                           "*",
                           c_SysInt_LHCLoggingDB_DataConsistency_Inconsistent,
                           c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery);
        dsAlertClasses = makeDynString(
                             "",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.",
                             getSystemName() + "_unSystemIntegrityAlarm_lvl2.");
        dbDiscreteNegation = makeDynBool(false, false, false);
        dsStateMatch = makeDynString("", "", "");
    } else {
        DebugTN("Invalid alert dp name: " + sSysAlarmDpe);
        return;
    }

    // set discrete alarm config
    dyn_mixed alarmObject;
    fwAlertConfig_objectCreateDiscrete(
        alarmObject,
        dsAlertTexts,
        dsAlertMatch,
        dsAlertClasses,
        "",
        makeDynString(""),
        "",
        false,
        dbDiscreteNegation,
        "",
        dsStateMatch,
        dsExInfo,
        false,
        false,
        "",
        false); //exception info returned here
    if (dynlen(dsExInfo)) { return;}
    //set the alarm to the dpe
    fwAlertConfig_objectSet(sSysAlarmDpe, alarmObject, dsExInfo);
    if (dynlen(dsExInfo)) { return;}
    //activate it
    fwAlertConfig_activate(sSysAlarmDpe, dsExInfo);
    if (dynlen(dsExInfo)) { return;}
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_manageMonitoringThread
/**
* This function is used to handle the monitorig thread which does the checking
*
* @param command         input, string: "start" or "stop" the thread
*/
private void unSystemIntegrity_LHCLoggingDB_manageMonitoringThread(string command)
{
    switch (command) {
        case "start": {
            // start only if thread is not running yet
            if (g_SysInt_LHCLoggingDB_monitoringThreadId == 0) {
                g_SysInt_LHCLoggingDB_monitoringThreadId = startThread("unSystemIntegrity_LHCLoggingDB_checkLogging");
            }
            break;
        }
        case "stop": {
            // stop thread only if necessary
            if (g_SysInt_LHCLoggingDB_monitoringThreadId != 0) {
                bool bStop = true;
                dyn_string dsSysIntAlarmDps = dpNames(c_unSystemAlarm_dpPattern + LHCLoggingDB_pattern + "*", c_unSystemAlarm_dpType);
                int iDsSysIntAlarmDpsLen = dynlen(dsSysIntAlarmDps);
                for (int i = 1; i <= iDsSysIntAlarmDpsLen; i++) {
                    bool bEnable;
                    dpGet(dsSysIntAlarmDps[i] + ".enabled", bEnable);
                    if (bEnable) {
                        bStop = false;
                    }
                }
                if (bStop) {
                    if (stopThread(g_SysInt_LHCLoggingDB_monitoringThreadId) < 0) {
                        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorStoppingThread);
                    } else {
                        g_SysInt_LHCLoggingDB_monitoringThreadId = 0;
                    }
                }
            }
            break;
        }
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_checkLogging
/**
* This is the function that actually does the system integrity checking in a periodic cycle
*
* Following checks are done:
* ----------------------------------------------------------
* 1. unSystemIntegrity_LHCLoggingDB_checkTransferJobs
* 2. unSystemIntegrity_LHCLoggingDB_checkTransferHistory
* 3. unSystemIntegrity_LHCLoggingDB_checkTestDatapoints
* 4. unSystemIntegrity_LHCLoggingDB_checkDataConsistency
*
* The result of every check is written to a dedicated alarm datapoint
* Checks are only executed if the alarm datapoint is enabled
*/
private void unSystemIntegrity_LHCLoggingDB_checkLogging()
{
    // Reconnect to database if db settings change
    string sSettingsDpe;
    if (unLHCLoggingDBSettings_getSettingsDpeName(sSettingsDpe) == false) {
        DebugTN("Settings Datapoint not found");
        return;
    }
    dpConnect("unSystemIntegrity_LHCLoggingDB_dbSettingsChangedCB", false, sSettingsDpe);
    dyn_errClass err = getLastError();
    if (dynlen(err) > 0) {
        DebugTN("Error on dpConnect: ", err);
        return;
    }

    string sAlarmDp;
    bool bAlarmEnabled;
    unsigned uResult;

    while (1) {
        int iErrorCode = 0;

        // 1. check status of transfer jobs
        sAlarmDp = c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_TransferJobs;
        if (dpExists(sAlarmDp)) {
            dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
            if (bAlarmEnabled) {
                uResult = unSystemIntegrity_LHCLoggingDB_checkTransferJobs(g_SysInt_LHCLoggingDB_mySystemDbSchema);
                dpSetWait(sAlarmDp + ".alarm", uResult);
            }
        }

        // 2. check transfer history of all registered elements
        sAlarmDp = c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_TransferHistory;
        if (dpExists(sAlarmDp)) {
            dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
            if (bAlarmEnabled) {
                uResult = unSystemIntegrity_LHCLoggingDB_checkTransferHistory(g_SysInt_LHCLoggingDB_mySystemDbSchema);
                dpSetWait(sAlarmDp + ".alarm", uResult);
            }
        }

        // 3. check transfer history of test datapoints
        sAlarmDp = c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_TestDatapoints;
        if (dpExists(sAlarmDp)) {
            dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
            if (bAlarmEnabled) {
                uResult = unSystemIntegrity_LHCLoggingDB_checkTestDatapoints();
                dpSetWait(sAlarmDp + ".alarm", uResult);
            }
        }

        // 4. Check if all registered elements have an entry in ELEMENTS table
        sAlarmDp = c_unSystemAlarm_dpPattern + c_SysInt_LHCLoggingDB_IntegrityClass_DataConsistency;
        if (dpExists(sAlarmDp)) {
            dpGet(sAlarmDp + ".enabled", bAlarmEnabled);
            if (bAlarmEnabled) {
                uResult = unSystemIntegrity_LHCLoggingDB_checkDataConsistency(g_SysInt_LHCLoggingDB_mySystemDbSchema);
                dpSetWait(sAlarmDp + ".alarm", uResult);
            }
        }

        delay(g_SysInt_LHCLoggingDB_loopDelay);
    }
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_dbSettingsChangedCB
/**
* Reconnect to db, if settings change
*
* @param sDpe        input, not used
* @param sValue      input, not used
*/
private void unSystemIntegrity_LHCLoggingDB_dbSettingsChangedCB(string sDpe, string sValue)
{
    unSystemIntegrity_LHCLoggingDB_initDabatase();
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_checkTransferJobs
/**
* Check current status for all transfer jobs related to sDbSchema
*
* @param sDbSchema        input, database schema for which we want to check the transfer jobs
*
* @return uReturnCode     output, Statuscode
*/
public unsigned unSystemIntegrity_LHCLoggingDB_checkTransferJobs(string sDbSchema)
{
    unsigned uReturnCode = c_SysInt_LHCLoggingDB_TransferJobs_OK;

    dyn_string dsExInfo;

    time tDbSysTimeUtc = unLHCLoggingDBSql_getDbSysTimeUtc(dsExInfo);
    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TransferJobs_ErrorQuery; }

    dyn_dyn_mixed ddmJobStatus = unLHCLoggingDBSql_getTransferJobStatus(sDbSchema + ".%", dsExInfo);
    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TransferJobs_ErrorQuery; }

    int iDdmJobStatusLen = dynlen(ddmJobStatus);
    for (int i = 1; i <= iDdmJobStatusLen; i++) {
        if (ddmJobStatus[i][2] != "TRUE") {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TransferJobs_ErrorDisabled) {
                uReturnCode = c_SysInt_LHCLoggingDB_TransferJobs_ErrorDisabled;
            }
            continue;
        }

        time tLastJobStart = ddmJobStatus[i][4];
        if (tDbSysTimeUtc - tLastJobStart > g_SysInt_LHCLoggingDB_maxAgeTransferJobs) {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TransferJobs_ErrorLagging) {
                uReturnCode = c_SysInt_LHCLoggingDB_TransferJobs_ErrorLagging;
            }
            continue;
        }
    }

    return uReturnCode;
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_checkTransferHistory
/**
* Check current status of all log variables from a specific database schema
*
* @param sDbSchema       input, DB schema to check
*
* @return uReturnCode    output, Statuscode with highest priority
*/
public unsigned unSystemIntegrity_LHCLoggingDB_checkTransferHistory(const string sDbSchema)
{
    dyn_string dsExInfo;
    unsigned uReturnCode = c_SysInt_LHCLoggingDB_TransferHistory_OK;

    time tDbSysTimeUtc = unLHCLoggingDBSql_getDbSysTimeUtc(dsExInfo);
    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TransferHistory_ErrorQuery; }

    dyn_dyn_mixed ddmLvsInfo = unLHCLoggingDBSql_getLvsInfo(sDbSchema, dsExInfo);
    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TransferHistory_ErrorQuery; }

    int iDdmLvsInfoLen = dynlen(ddmLvsInfo);
    for (int i = 1; i <= iDdmLvsInfoLen; i++) {
        // Ensure that VARIABLE_NAME has a CREATED_UTC_STAMP, if not return ERROR
        if (ddmLvsInfo[i][2] == "") {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TransferHistory_ErrorRegistration) {
                uReturnCode = c_SysInt_LHCLoggingDB_TransferHistory_ErrorRegistration;
            }
            continue;
        }

        // Ensure that all CREATED_UTC_STAMP are up to date (not older than TIME_DIFF_MAX), if not return ERROR
        time tCreatedUtc = ddmLvsInfo[i][2];
        if (tDbSysTimeUtc - tCreatedUtc > g_SysInt_LHCLoggingDB_maxAgeTransferHistory) {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TransferHistory_ErrorTransfer) {
                uReturnCode = c_SysInt_LHCLoggingDB_TransferHistory_ErrorTransfer;
            }
            continue;
        }

        // Ensure that VARIABLE_NAME has a NO_RECS_LOGGED and MAX_ROWS, if not return ERROR
        if (ddmLvsInfo[i][4] == "" || ddmLvsInfo[i][5] == "") {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TransferHistory_ErrorRegistration) {
                uReturnCode = c_SysInt_LHCLoggingDB_TransferHistory_ErrorRegistration;
            }
            continue;
        }

        // Ensure that data transfer is not lagging
        unsigned uNoRecsLogged = ddmLvsInfo[i][4];
        unsigned uNoRecsMax = ddmLvsInfo[i][5];
        if (uNoRecsLogged == uNoRecsMax) {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TransferHistory_WarningLagging) {
                uReturnCode = c_SysInt_LHCLoggingDB_TransferHistory_WarningLagging;
            }
            continue;
        }
    }
    return uReturnCode;
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_checkTestDatapoints
/**
* Check current status for all test datapoints
*
* @return uReturnCode     output, Statuscode
*/
public unsigned unSystemIntegrity_LHCLoggingDB_checkTestDatapoints()
{
    // Get variable names of test data points
    dyn_string dsVariableName = unSystemIntegrity_LHCLoggingDB_getTestDpVarNames();

    dyn_string dsExInfo;
    unsigned uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_OK;

    time tDbSysTimeUtc = unLHCLoggingDBSql_getDbSysTimeUtc(dsExInfo);
    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery; }

    int iDsVariableNameLen = dynlen(dsVariableName);
    for (int i = 1; i <= iDsVariableNameLen; i++) {
        // Check if element is registered, if no, set registration error
        bool bRegistered = unLHCLoggingDBSql_isElementRegistered(dsVariableName[i], dsExInfo);
        if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery; }
        if (bRegistered == false) {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration) {
                uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration;
            }
            continue;
        }

        // Check transfer history, if there are no values we are initialising
        dyn_dyn_mixed ddmTestDpLvsData = unLHCLoggingDBSql_getLvsForTestDpe(dsVariableName[i], dsExInfo);
        if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_TestDatapoints_ErrorQuery; }

        // Do we have 3 results ?
        if (dynlen(ddmTestDpLvsData) != 3) {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TestDatapoints_Initialising) {
                uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_Initialising;
            }
            continue;
        }

        // Do all results have an LAST_LOGGED_UTC timestamp ?
        if (ddmTestDpLvsData[1][1] == "" || ddmTestDpLvsData[2][1] == "" || ddmTestDpLvsData[3][1] == "" ||
                ddmTestDpLvsData[1][2] == "" || ddmTestDpLvsData[2][2] == "" || ddmTestDpLvsData[3][2] == "") {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration) {
                uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration;
            }
            continue;
        }

        time tCreatedUtc = ddmTestDpLvsData[1][1];
        if (tDbSysTimeUtc - tCreatedUtc > g_SysInt_LHCLoggingDB_maxAgeTransferHistory) {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration) {
                uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_ErrorRegistration;
            }
            continue;
        }

        time t1 = ddmTestDpLvsData[1][2];
        time t2 = ddmTestDpLvsData[2][2];
        time t3 = ddmTestDpLvsData[3][2];

        // Are the times increasing ?
        if (t1 > t2 && t2 > t3 && t1 > t3) {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TestDatapoints_OK) {
                uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_OK;
            }
            continue;
        }

        // Are the times at least changing ?
        if (t1 != t2 && t2 != t3 && t1 != t3) {
            if (uReturnCode < c_SysInt_LHCLoggingDB_TestDatapoints_Retransfering) {
                uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_Retransfering;
            }
            continue;
        }

        if (uReturnCode < c_SysInt_LHCLoggingDB_TestDatapoints_ErrorTransfer) {
            uReturnCode = c_SysInt_LHCLoggingDB_TestDatapoints_ErrorTransfer;
        }
    }

    return uReturnCode;
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_checkDataConsistency
/**
* Check data consistecy between registered variables in database (table: META_VARIABLES)
* and datapoints with an archive config in the PVSS project
*
* @param sDbSchema       input, DB schema to check
*
* @return uReturnCode    output, Statuscode
*/
public unsigned unSystemIntegrity_LHCLoggingDB_checkDataConsistency(string sDbSchema)
{
    dyn_string dsExInfo;
    dyn_dyn_mixed ddmData = unLHCLoggingDBSql_getInconsistentData(sDbSchema, dsExInfo);

    if (dynlen(dsExInfo) > 0) { return c_SysInt_LHCLoggingDB_DataConsistency_ErrorQuery; }
    if (dynlen(ddmData) <= 0) { return c_SysInt_LHCLoggingDB_DataConsistency_OK; }
    return c_SysInt_LHCLoggingDB_DataConsistency_Inconsistent;
}

//------------------------------------------------------------------------------------------------------------------------

// unSystemIntegrity_LHCLoggingDB_throwError
/**
* Write special error to message log
*
* @param severity     input, severity of the message
* @param code         input, error code to write
* @param sDetails     input, message details
*/
private void unSystemIntegrity_LHCLoggingDB_throwError(string severity = "ERROR", int code = 1, string sDetails = "")
{
    int prio = PRIO_WARNING; //seems that throwing a fatal error makes Pmon stop monitoring this manager????
    int type = ERR_CONTROL;

    switch (strtoupper(severity)) {
        case "INFO":
            prio = PRIO_INFO;
            break;
        case "WARNING":
            prio = PRIO_WARNING;
            break;
        case "ERROR":
            prio = PRIO_SEVERE;
            break;
    }

    errClass err = makeError("unSystemIntegrity_LHCLoggingDB", prio, type, code, sDetails);
    throwError(err);

    return;
}

//------------------------------------------------------------------------------------------------------------------------


// unLHCLoggingDB_doFullConsistencyCheck
/**
* Check consistency between config files and registrated data in db
*
* @param dsExInfo     output, exeption info
*/
private void unSystemIntegrity_LHCLoggingDB_doFullConsistencyCheck(dyn_string &dsExInfo)
{
    mapping mFile2Db;
    mFile2Db["VarName"] = makeDynString();
    mFile2Db["Exist"] = makeDynString();
    mFile2Db["Match"] = makeDynString();
    mFile2Db["Info"] = makeDynString();
    mFile2Db["ConfigFile"] = makeDynString();
    mFile2Db["Errors"] = 0;
    mapping mDb2File;
    mDb2File["VarName"] = makeDynString();
    mDb2File["Exist"] = makeDynString();
    mDb2File["Match"] = makeDynString();
    mDb2File["Info"] = makeDynString();
    mDb2File["Errors"] = 0;
    mapping mResult;
    mResult["Status"] = true;
    mResult["File2Db"] = mFile2Db;
    mResult["Db2File"] = mDb2File;

    // Get configuration data from database
    string sMySystemDbSchema = unLHCLoggingDBSettings_getMySystemDbSchema();

    // Get elements in MDB
    dyn_dyn_mixed ddmElements = unLHCLoggingDBSql_getElements(sMySystemDbSchema, dsExInfo);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorFullConsistencyCheck, "Elements");
        return;
    }
    dyn_string dsElementNameFromElements;
    if (dynlen(ddmElements[1]) > 0) {
        dsElementNameFromElements = ddmElements[1];
    }

    // Get registered variables
    dyn_dyn_mixed ddmMetaVars = unLHCLoggingDBSql_getLogVarInfo(sMySystemDbSchema, dsExInfo, true);
    if (dynlen(dsExInfo) > 0) {
        unSystemIntegrity_LHCLoggingDB_throwError("ERROR", c_SysInt_LHCLoggingDB_Catalogue_ErrorFullConsistencyCheck, "Meta Variables");
        return;
    }
    dyn_string dsVariableNameFromDb, dsElementNameFromDb, dsElementDescriptionFromDb, dsHierarchyFromDb;
    if (dynlen(ddmMetaVars) > 0) {
        dsVariableNameFromDb = ddmMetaVars[2];
        dsElementNameFromDb = ddmMetaVars[3];
        dsElementDescriptionFromDb = ddmMetaVars[15];
        dsHierarchyFromDb = ddmMetaVars[10];
    }

    // Get configuration data from files
    dyn_string dsImportFiles, dsConfigFiles, dsVarNameFromFile, dsDpeFromFile, dsHierarchyFromFile, dsDescriptionFromFile, dsDataCategoryFromFile;
    dpGet("unLHCLoggingDB_ImportFiles.", dsImportFiles);
    int iDsImportFilesLen = dynlen(dsImportFiles);
    for (int i = 1; i <= iDsImportFilesLen; i++) {
        dyn_string dsVarName, dsDpe, dsHierarchy, dsDescription, dsDataCategory, dsConfigFile;
        unLHCLoggingDBParser_getConfigFileData(dsImportFiles[i], dsVarName, dsDpe, dsHierarchy, dsDescription, dsDataCategory);

        int iDsVarNameLen = dynlen(dsVarName);
        for (int j = 1; j <= iDsVarNameLen; j++) {
            dsConfigFile[j] = dsImportFiles[i];
        }

        dynAppend(dsConfigFiles, dsConfigFile);
        dynAppend(dsVarNameFromFile, dsVarName);
        dynAppend(dsDpeFromFile, dsDpe);
        dynAppend(dsHierarchyFromFile, dsHierarchy);
        dynAppend(dsDescriptionFromFile, dsDescription);
        dynAppend(dsDataCategoryFromFile, dsDataCategory);
    }

    // Take test datapoints into account
    dyn_string dsTestDp = unSystemIntegrity_LHCLoggingDB_getTestDps();   
    int len = dynlen(dsTestDp);
    for (int i = 1; i <= len; i++) {
        string sDataCategory = unSystemIntegrity_LHCLoggingDB_getDataCategory(dsTestDp[i]);
        strreplace(sDataCategory, "/", "_");
        string sDpeName = dsTestDp[i] + c_SysInt_RDB_TestDpe_Value;
        string sVarName = strtoupper(unSystemIntegrity_LHCLoggingDB_getVarName(dsTestDp[i]));
        string sDescription = dpGetDescription(sDpeName);
        dynAppend(dsConfigFiles, "automatically created");
        dynAppend(dsVarNameFromFile, sVarName);
        dynAppend(dsDpeFromFile, sDpeName);
        dynAppend(dsHierarchyFromFile, "");
        dynAppend(dsDescriptionFromFile, sDescription);
        dynAppend(dsDataCategoryFromFile, sDataCategory);
    }

    // Check: FILE --> DB
    // =================  //
    int iDsVarNameFromFileLen = dynlen(dsVarNameFromFile);
    for (int i = 1; i <= iDsVarNameFromFileLen; i++) {
        mResult["File2Db"]["VarName"][i] = dsVarNameFromFile[i];
        mResult["File2Db"]["Exist"][i] = "NO";
        mResult["File2Db"]["Match"][i] = "NO";
        mResult["File2Db"]["Info"][i] = "";
        mResult["File2Db"]["ConfigFile"][i] = dsConfigFiles[i];

        // Check if Variable_Name exists in database
        int dbIdx = dynContains(dsVariableNameFromDb, dsVarNameFromFile[i]);
        if (dbIdx <= 0) {
            mResult["File2Db"]["Info"][i] = "Variable is not registered for logging!";
            mResult["File2Db"]["Errors"] += 1;
            continue;
        }
        mResult["File2Db"]["Exist"][i] = "YES";

        // Check if Dpe exists on system
        if (dsDpeFromFile[i] == "") {
            mResult["File2Db"]["Info"][i] = "DP does not exist on system!";
            mResult["File2Db"]["Errors"] += 1;
            continue;
        }

        // Check if Dpe exists in elements table
        if (dynContains(dsElementNameFromElements, dsDpeFromFile[i]) <= 0) {
            mResult["File2Db"]["Info"][i] = "DP does not exist in MDB!";
            mResult["File2Db"]["Errors"] += 1;
            continue;
        }

        // Check if Dpe from config file matches Element_Name from db
        if (dsDpeFromFile[i] != dsElementNameFromDb[dbIdx]) {
            mResult["File2Db"]["Info"][i] = "DP does not match with registered DP";
            mResult["File2Db"]["Errors"] += 1;
            continue;
        }

        //Check if description from config file matches description in db
        if (dsDescriptionFromFile[i] != dsElementDescriptionFromDb[dbIdx])  {
            mResult["File2Db"]["Info"][i] = "Description does not match!";
            mResult["File2Db"]["Errors"] += 1;
            continue;
        }

        //Check if hierarchy from config file matches hierarchy in db
        if (dsHierarchyFromFile[i] != dsHierarchyFromDb[dbIdx]) {
            mResult["File2Db"]["Info"][i] = "Hierarchy does not match!";
            mResult["File2Db"]["Errors"] += 1;
            continue;
        }

        mResult["File2Db"]["Match"][i] = "YES";
        mResult["File2Db"]["Info"][i] = "";
    }

    // Check: DB --> FILE
    // =================  //
    int iDsVariableNameFromDbLen = dynlen(dsVariableNameFromDb);
    for (int i = 1; i <= iDsVariableNameFromDbLen; i++) {
        mResult["Db2File"]["VarName"][i] = dsVariableNameFromDb[i];
        mResult["Db2File"]["Exist"][i] = "NO";
        mResult["Db2File"]["Match"][i] = "NO";
        mResult["Db2File"]["Info"][i] = "";

        // Check if Variable_Name from database is found in config file
        int fileIdx = dynContains(dsVarNameFromFile, dsVariableNameFromDb[i]);
        if (fileIdx <= 0) {
            mResult["Db2File"]["Info"][i] = "Variable not found in config file!";
            mResult["Db2File"]["Errors"] += 1;
            continue;
        }
        mResult["Db2File"]["Exist"][i] = "YES";

        // Check if Element_Name for this Variable_Name exists in database
        if (dsElementNameFromDb[i] == "") {
            mResult["Db2File"]["Info"][i] = "Variable has no associated DP in MDB!";
            mResult["Db2File"]["Errors"] += 1;
            continue;
        }

        // Check if Element_Name exists on this system
        if (!dpExists(dsElementNameFromDb[i])) {
            mResult["Db2File"]["Info"][i] = "DP does not exist on system!";
            mResult["Db2File"]["Errors"] += 1;
            continue;
        }

        // Check if Dpe from DB matches dp from config file
        if (dsElementNameFromDb[i] != dsDpeFromFile[fileIdx]) {
            mResult["Db2File"]["Info"][i] = "DP does not match with DP in config file!";
            mResult["Db2File"]["Errors"] += 1;
            continue;
        }

        //Check if description from config file matches description in db
        if (dsElementDescriptionFromDb[i] != dsDescriptionFromFile[fileIdx])  {
            mResult["Db2File"]["Info"][i] = "Description does not match!";
            mResult["Db2File"]["Errors"] += 1;
            continue;
        }

        //Check if hierarchy from config file matches hierarchy in db
        if (dsHierarchyFromDb[i] != dsHierarchyFromFile[fileIdx]) {
            mResult["Db2File"]["Info"][i] = "Hierarchy does not match!";
            mResult["Db2File"]["Errors"] += 1;
            continue;
        }

        mResult["Db2File"]["Match"][i] = "YES";
        mResult["Db2File"]["Info"][i] = "";
    }

    mResult["Status"] = ((mResult["File2Db"]["Errors"] + mResult["Db2File"]["Errors"]) == 0);

    // Write results into dp
    const string sResultDP = "unLHCLoggingDB_ConsistencyCheck";
    dpSetWait(sResultDP + ".File2Db.VarName", mResult["File2Db"]["VarName"],
              sResultDP + ".File2Db.Exist", mResult["File2Db"]["Exist"],
              sResultDP + ".File2Db.Match", mResult["File2Db"]["Match"],
              sResultDP + ".File2Db.Info", mResult["File2Db"]["Info"],
              sResultDP + ".File2Db.ConfigFile", mResult["File2Db"]["ConfigFile"],
              sResultDP + ".File2Db.Errors", mResult["File2Db"]["Errors"],
              sResultDP + ".Db2File.VarName", mResult["Db2File"]["VarName"],
              sResultDP + ".Db2File.Exist", mResult["Db2File"]["Exist"],
              sResultDP + ".Db2File.Match", mResult["Db2File"]["Match"],
              sResultDP + ".Db2File.Info", mResult["Db2File"]["Info"],
              sResultDP + ".Db2File.Errors", mResult["Db2File"]["Errors"],
              sResultDP + ".Status", mResult["Status"] );
}

