main()
{
  string sDbName, sDbUser, sDbPass;  
  
  // Get database settings from datapoint    
  if(getActiveDbConfig(sDbName, sDbUser, sDbPass)==false) 
    return;
  
  // Establish database connection
  if (unLHCLoggingDBSql_openDbConnection(sDbName, sDbUser, sDbPass) == false) 
    return; 
  
  // Get DB schema of current system 
  string sMySystemDbSchema = getMySystemDbSchema();

  
  // Get Rdb source of current system
  string sMySystemRdbSource;
  if(getRdbSource(sMySystemDbSchema, sMySystemRdbSource) == false) 
    return;  
  
  const unsigned T_LOOP_DELAY = 300; // [seconds]
      
  while (1)
  {  
    DebugTN("Checking 'QPS_81_DL1DE.QPSHISTORY' Return code = " + unLHCLoggingDB_checkLvs("QPS_81_DL1DE.QPSHISTORY")); 
    
    DebugTN("Checking variable 'MB.C22L1:U_DIODE_RB' Return code = " + unLHCLoggingDB_checkTestDPs("MB.C22L1:U_DIODE_RB")); 
    DebugTN("Checking variable 'DQQDL.A25L1:ST_LOOP_CLOSED' Return code = " + unLHCLoggingDB_checkTestDPs("DQQDL.A25L1:ST_LOOP_CLOSED")); 
    DebugTN("Checking variable 'DQAMG.B24L1.RQF.A81:ST_LOGGING' Return code = " + unLHCLoggingDB_checkTestDPs("DQAMG.B24L1.RQF.A81:ST_LOGGING")); 
        
    delay(T_LOOP_DELAY);
  }
}


// TODO: think about this solution, maybe change, that RDB_SOURCE will be the return value
/**
* Get RdbSource from DbSchema
*
* @param   sDbSchema    input,  Database schema
* @param   sRdbSource   output, Rdb Source = Schema + Tablename + "History"

* @return               output, true=success,false=error 
*/
bool getRdbSource(string sDbSchema, string &sRdbSource)
{
  string sql = "select distinct RDB_SOURCE from RDB2LHC.META_VARIABLES where RDB_SOURCE like '"+sDbSchema+".%HISTORY'";
  
  dyn_string dsExInfo;
  dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryData(sql, dsExInfo);
  
  if(dynlen(sqlResult) != 1) { 
    DebugTN("Query error"); 
    return false; 
  };
  
  sRdbSource = sqlResult[1][1];  
  
  return true;  
}





dyn_dyn_mixed g_ddmRdbChangesMemory; 
dyn_dyn_mixed g_ddmLvsChangesMemory; 

bool doCausalityCheck(string sRdbSource)    
{ 
  
  time tCurrent = getCurrentTime()-1*24*60*60; 

  while(1)
  {     
    DebugTN("tCurrent: ", tCurrent);
    
    const unsigned tMaxError = 0;      // Max. Time for response after valuechange in RDB Archive 10min
    const unsigned tQueryInterval = 3600; // Time period for query    
       
    time tEnd = tCurrent - tMaxError;
    time tStart = tEnd - tQueryInterval;     
    
    dyn_string dsVarsToCheck = getRdbArchiveChanges (sRdbSource, tStart, tEnd);  
    
    dyn_string dsErrors = getLogVariableStatusChanges (dsVarsToCheck, sRdbSource, tStart, tCurrent);
    
    DebugTN(dsErrors);    

    tCurrent += tQueryInterval+tMaxError + 0; 
    
    if(tCurrent > getCurrentTime())
      delay(tQueryInterval+tMaxError+0);    
  }
                
   
  return false;  
}


dyn_string getRdbArchiveChanges (string sRdbSource, time tStart, time tEnd)
{
  dyn_string dsVarsToCheck;
  
  dyn_dyn_mixed ddmRdbChangesNew = queryRdbArchive(sRdbSource, tStart, tEnd);
  for(int i=1; i<=dynlen(ddmRdbChangesNew);i++)
  {
    // Does Memory already contain new VARIABLE_ID ?  
    int iId = dynContains(getDynString(g_ddmRdbChangesMemory,1),ddmRdbChangesNew[i][1]);       
    if(iId > 0) 
    {
      // Yes
      
      // Did the VALUE_NUMBER change ?
      if(g_ddmRdbChangesMemory[iId][2] != ddmRdbChangesNew[i][2])
      {
        // Yes, mark this varId to be checked later in LVS
        dynAppend(dsVarsToCheck, ddmRdbChangesNew[i][1]);         
        
        // Remember new VALUE_NUMBER 
        g_ddmRdbChangesMemory[iId][2] = ddmRdbChangesNew[i][2];        
      }      
    }
    else if(iId == 0) 
    {
      //  No
      
      // Remember new VARIABLE_ID and VALUE_NUMBER
      dynAppend(g_ddmRdbChangesMemory, makeDynMixed(ddmRdbChangesNew[i][1], ddmRdbChangesNew[i][2]));    
    }
    else 
    {
      DebugTN("Error occured");      
    }
  }   
    
  DebugTN("Vars in memory: " + dynlen(g_ddmRdbChangesMemory) + "  Vars to check: " + dynlen(dsVarsToCheck)); 
  return dsVarsToCheck;  
}

dyn_string getLogVariableStatusChanges (dyn_string dsVarsToCheck, string sRdbSource, time tStart, time tEnd)
{
  dyn_string dsErrors;
  
  dyn_dyn_mixed ddmLvsChangesNew = queryLogVariableStatus(sRdbSource, tStart, tEnd);
  for(int i=1; i<=dynlen(dsVarsToCheck);i++)
  {
    // Do new changes contain VARIABLE_ID where timestamp changed earlier?  
    int iId = dynContains(getDynString(ddmLvsChangesNew,1),dsVarsToCheck[i]);       
    if(iId > 0) 
    { 
      // Yes, LvsChangesNew contains VARIABLE_ID
      
      // Does LvsChangesMemory contain VARIABLE_ID
      int iIdMem = dynContains(getDynString(g_ddmLvsChangesMemory,1),ddmLvsChangesNew[iId][1]); 
      if(iIdMem > 0) 
      {
        // Yes, LvsChangesMemory contains VARIABLE_ID        
        
        // Did the LAST_LOGGED_UTC_STAMP change ?
        if(g_ddmLvsChangesMemory[iIdMem][2] != ddmLvsChangesNew[iId][2]) 
        {
          // Yes
          
          // Remember new LAST_LOGGED_UTC_STAMP
          g_ddmLvsChangesMemory[iIdMem][2] = ddmLvsChangesNew[iId][2];         
          
        }
        else
        {
          //  No  
          
         dynAppend(dsErrors, "ERROR 1: Element " + dsVarsToCheck[i] + " not transferred to LDB");         
        }
      }
      else if(iIdMem == 0)
      {
        // No, LvsChangesMemory does not contain VARIABLE_ID
        
        // Append VARIABLE_ID and LAST_LOGGED_UTC_STAMP
        dynAppend(g_ddmLvsChangesMemory, makeDynMixed(ddmLvsChangesNew[iId][1], ddmLvsChangesNew[iId][2]));    
      }
      else 
      {
        DebugTN("Error occured");      
      }
  
    } 
    else if(iId == 0) 
    {
      // NO, LvsChangesNew does not contain VARIABLE_ID
      
      dynAppend(dsErrors, "ERROR 2: Element " + dsVarsToCheck[i] + " not transferred to LDB");
    } 
    else 
    {
      DebugTN("Error occured");      
    }    
  } 
  
  DebugTN("Vars in memory: " + dynlen(g_ddmLvsChangesMemory) + "  Vars to check: " + dynlen(dsVarsToCheck)); 
  return dsErrors;  
}

dyn_dyn_mixed queryRdbArchive(string sRdbSource, time tStart, time tEnd)
{   
  string sql = "select to_char(H.ELEMENT_ID), to_char(H.VALUE_NUMBER) from "
               "( "
               " select HI.ELEMENT_ID, HI.VALUE_NUMBER, row_number() over ( partition by HI.ELEMENT_ID order by HI.TS desc nulls last) ROW_NUM "
               "  from "+sRdbSource+" HI join RDB2LHC.META_VARIABLES MV on MV.VARIABLE_ID = HI.ELEMENT_ID  "
               " where HI.TS between TO_TIMESTAMP(:t1,'YYYY.MM.DD HH24:MI:SS.FF') and TO_TIMESTAMP(:t2,'YYYY.MM.DD HH24:MI:SS.FF') "
               ") H "
              "where ROW_NUM = 1 ";
  
  mapping params;
  params[":t1"]=formatTimeUTC("%Y.%m.%d %H:%M:%S.", tStart, "%03d"); 
  params[":t2"]=formatTimeUTC("%Y.%m.%d %H:%M:%S.", tEnd, "%03d");   
  
  dyn_string exceptionInfo;
  dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryDataWithParMapping(sql, params, exceptionInfo);
  
  return sqlResult;  
}


dyn_dyn_mixed queryLogVariableStatus(string sRdbSource, time tStart, time tEnd)
{  
  string sql = "select to_char(mv.variable_id), max(lvs.last_logged_utc_stamp) from RDB2LHC.META_VARIABLES mv "
               " left outer join RDB2LHC.LOG_VARIABLE_STATUS lvs on mv.variable_id = lvs.variable_id "
               "where RDB_SOURCE = :RdbSource and lvs.created_utc_stamp between TO_TIMESTAMP(:t1,'YYYY.MM.DD HH24:MI:SS.FF') and TO_TIMESTAMP(:t2,'YYYY.MM.DD HH24:MI:SS.FF') "  
               " group by mv.variable_id ";  
  
  mapping params;
  params[":RdbSource"]=sRdbSource;
  params[":t1"]=formatTimeUTC("%Y.%m.%d %H:%M:%S.", tStart, "%03d"); 
  params[":t2"]=formatTimeUTC("%Y.%m.%d %H:%M:%S.", tEnd, "%03d");   
  
  dyn_string exceptionInfo;
  dyn_dyn_mixed sqlResult = unLHCLoggingDBSql_queryDataWithParMapping(sql, params, exceptionInfo);
  
  return sqlResult;  
}
























const unsigned COMPARE_MODE_EQUAL = 0;
const unsigned COMPARE_MODE_EQUAL_MIN_1 = 1;

bool compareNoOfRecs(ulong ulRecsFromLVS, ulong ulRecsFromRDB, unsigned uCompareMode)
{
  switch(uCompareMode)
  {
    case COMPARE_MODE_EQUAL:   
      return (ulRecsFromLVS==ulRecsFromRDB ? true : false);        
    case COMPARE_MODE_EQUAL_MIN_1:   
      return (ulRecsFromLVS==(ulRecsFromRDB-1) ? true : false);          
    default:  
      return (false);    
  }
 
  return false;  
}


/**
* checkLoggingIntegrity algorithm
*/
void checkLoggingIntegrity() // TODO: make VARIABLE_ID an argument of this function, or make RDB_SOURCE an argument of this function?
{   
bool bInit=false;
bool bDummy=false;

unsigned uCompareMode = COMPARE_MODE_EQUAL;

  
string sVariableId="144137865754870618";
string sRdbSource="QPS_12_DR1HI.QPSHISTORY";  // TODO: calculate RDSB Source from sVAriableID (select RDB_SOURCE from RDB2LHC.META_VARIABLES where VARIABLE_ID=sVariableId ... )

time tStart = makeTime(2014, 5, 26, 05, 3, 37, 200);
time tEnd;
const long tInterval = 3600;
time tFirstCheckedMin; 
time tLastCheckedMax;

ulong ulLVSRecs;  
ulong ulRDBRecs; 

int iErrorCounter;

string sDebug;

  while (1)
  { 
    tEnd = tStart + tInterval; 
    
    if(tEnd >= getCurrentTime()) {
      DebugTN("Waiting some time: ");
      delay(tInterval);      
    }  
    
    getLogVarStatus(sVariableId, tStart, tEnd, tFirstCheckedMin, tLastCheckedMax, ulLVSRecs);

    if(ulLVSRecs > 0)
    {
      // Entry found between timestamps, check if the number of records match with records in RDB      
      
      if(bInit==false) 
      {
        bInit=true;  
        
        uCompareMode = COMPARE_MODE_EQUAL;
        
        // Get number of records from MDB within this timerange
        getRdbDataCount(sVariableId, sRdbSource, tFirstCheckedMin, tLastCheckedMax, ulRDBRecs); 
       
        sDebug = "From " + formatTime("%Y.%m.%d %H:%M:%S.", tFirstCheckedMin, "%03d") + 
                 " to " + formatTime("%Y.%m.%d %H:%M:%S.", tLastCheckedMax, "%03d") + 
                 " Recs LVS " + ulLVSRecs + 
                 " Recs RDB " + ulRDBRecs;                 
      }
      else
      {
        // Get number of records from MDB within this timerange
        getRdbDataCount(sVariableId, sRdbSource, tStart, tLastCheckedMax, ulRDBRecs); 
       
        sDebug = "From " + formatTime("%Y.%m.%d %H:%M:%S.", tStart, "%03d") + 
                 " to " + formatTime("%Y.%m.%d %H:%M:%S.", tLastCheckedMax, "%03d") + 
                 " Recs LVS " + ulLVSRecs + 
                 " Recs RDB " + ulRDBRecs;                        
      }
      
      if (true == compareNoOfRecs(ulLVSRecs, ulRDBRecs, uCompareMode)) { 
        sDebug += " STATUS: OK";
      }
      else { 
        sDebug += " STATUS: ERROR" ;  
      }    
      DebugTN(sDebug);  
      
      uCompareMode = COMPARE_MODE_EQUAL_MIN_1;
      
      tStart = tLastCheckedMax;       
      iErrorCounter = 0;
    }
    else        
    {             
      // Look for overlapping chunks in LVS       
      checkForChunksThatOverlap(sVariableId, tStart, tEnd, ulLVSRecs);  
      
      // Look for data in RDB      
      getRdbDataCount(sVariableId, sRdbSource, tStart, tEnd, ulRDBRecs); 
        
      sDebug = "From " + formatTime("%Y.%m.%d %H:%M:%S.", tStart, "%03d") + 
               " to " + formatTime("%Y.%m.%d %H:%M:%S.", tEnd, "%03d") + 
               " Recs LVS " + 0 + 
               " Recs RDB " + ulRDBRecs; 
      
      if(ulRDBRecs == ulLVSRecs) 
      {
        sDebug += " STATUS: OK" ;      
        DebugTN(sDebug);  
          
        tStart = tEnd;       
        iErrorCounter = 0;             
      }
      else
      {          
        if((++iErrorCounter) >= 5) 
        {          
          sDebug += " STATUS: ERROR" ;      
          DebugTN(sDebug);  
            
          tStart = tEnd;       
          iErrorCounter = 0;            
        } 
          
        delay(300);       
      } 
      
      uCompareMode = COMPARE_MODE_EQUAL;          
    }   
    delay(1);    
  }  
}


bool getLogVarStatus(string sVariableId, time tStart, time tEnd, time &tMinFirstChecked, time &tMaxLastChecked, ulong &ulRecs)
{
  string sql = "select to_char(min(first_checked_utc_stamp),'YYYY.MM.DD HH24:MI:SS.FF'), to_char(max(last_checked_utc_stamp),'YYYY.MM.DD HH24:MI:SS.FF'), sum(no_recs_logged) " 
               " from RDB2LHC.log_variable_status " 
               "  where variable_id = :varId and "+
               "        first_checked_utc_stamp >= TO_TIMESTAMP(:t1,'YYYY.MM.DD HH24:MI:SS.FF') and "+
               "        last_checked_utc_stamp <= TO_TIMESTAMP(:t2,'YYYY.MM.DD HH24:MI:SS.FF') ";
  
  mapping params;
  params[":varId"]=sVariableId;
  params[":t1"]=formatTime("%Y.%m.%d %H:%M:%S.", tStart, "%03d"); 
  params[":t2"]=formatTime("%Y.%m.%d %H:%M:%S.", tEnd, "%03d"); 
  
  dyn_string exceptionInfo;
  dyn_dyn_mixed data = unLHCLoggingDBSql_queryDataWithParMapping(sql, params, exceptionInfo);

  tMinFirstChecked = data[1][1];
  tMaxLastChecked = data[1][2];
  ulRecs = (data[1][3]=="" ? 0 : data[1][3]);

  return true;   
}

bool getRdbDataCount(string sVariableId, string sRdbSource, time tStart, time tEnd, ulong &ulRecs)
{
  string sql = "select count(1) " +
               " from " + sRdbSource +
               "  where element_id = :varId and " +
               "        TS >= TO_TIMESTAMP(:t1,'YYYY.MM.DD HH24:MI:SS.FF') and " +
               "        TS <= TO_TIMESTAMP(:t2,'YYYY.MM.DD HH24:MI:SS.FF') ";  

  mapping params;
  params[":varId"]=sVariableId;
  params[":t1"]=formatTime("%Y.%m.%d %H:%M:%S.", tStart, "%03d"); 
  params[":t2"]=formatTime("%Y.%m.%d %H:%M:%S.", tEnd, "%03d"); 
  
  dyn_string exceptionInfo;
  dyn_dyn_mixed data = unLHCLoggingDBSql_queryDataWithParMapping(sql, params, exceptionInfo);
  
  ulRecs = (data[1][1]=="" ? 0 : data[1][1]);
  
  return true;  
} 


bool checkForChunksThatOverlap(string sVariableId, time &tStart, time &tEnd, ulong &ulLVSRecs)
{ 
/*
 string sql1 = "select to_char(max(last_checked_utc_stamp),'YYYY.MM.DD HH24:MI:SS.FF') " +  
               " from RDB2LHC.log_variable_status  " + 
               " where variable_id=:varId " +  
	             " and ( " +
			         "  (first_checked_utc_stamp < TO_TIMESTAMP(:t1,'YYYY.MM.DD HH24:MI:SS.FF') " +
			         "  and last_checked_utc_stamp >= TO_TIMESTAMP(:t1,'YYYY.MM.DD HH24:MI:SS.FF')) " +  			
		           " ) " ;*/  
  
  string sql1 = "select to_char(max(last_checked_utc_stamp),'YYYY.MM.DD HH24:MI:SS.FF') " +  
               " from RDB2LHC.log_variable_status  " + 
               " where variable_id='"+sVariableId+"' " +  
	             " and ( " +
			         "  (first_checked_utc_stamp < TO_TIMESTAMP('"+formatTime("%Y.%m.%d %H:%M:%S.", tStart, "%03d")+"','YYYY.MM.DD HH24:MI:SS.FF') " +
			         "  and last_checked_utc_stamp >= TO_TIMESTAMP('"+formatTime("%Y.%m.%d %H:%M:%S.", tStart, "%03d")+"','YYYY.MM.DD HH24:MI:SS.FF')) " +  			
		           " ) " ;
 
  string sql2 = "select to_char(min(first_checked_utc_stamp),'YYYY.MM.DD HH24:MI:SS.FF') " +
               " from RDB2LHC.log_variable_status   " +
               " where variable_id='"+sVariableId+"' " +
	             " and ( " +
			         "  (first_checked_utc_stamp <= TO_TIMESTAMP('"+formatTime("%Y.%m.%d %H:%M:%S.", tEnd, "%03d")+"','YYYY.MM.DD HH24:MI:SS.FF') " +
			         "  and last_checked_utc_stamp > TO_TIMESTAMP('"+formatTime("%Y.%m.%d %H:%M:%S.", tEnd, "%03d")+"','YYYY.MM.DD HH24:MI:SS.FF')) "	+		
		           " ) " ; 
  
  ulLVSRecs = 0;
  
//  mapping params1;
//  params1[":varId"]=sVariableId;
//  params1[":t1"]=formatTime("%Y.%m.%d %H:%M:%S.", tStart, "%03d"); 
  
//  dyn_string exceptionInfo;
//  dyn_dyn_mixed data = unLHCLoggingDBSql_queryDataWithParMapping(sql1, params1, exceptionInfo); 
 
  dyn_string dsExInfo;  
  
  dyn_dyn_mixed data = unLHCLoggingDBSql_queryData(sql1, dsExInfo);   
  if(data[1][1]!="") {
    tStart = data[1][1];  
    DebugTN("Overlapping: New tStart", formatTime("%Y.%m.%d %H:%M:%S.", tStart, "%03d"));
    ulLVSRecs++;
  }    

  dyn_dyn_mixed data2 = unLHCLoggingDBSql_queryData(sql2, dsExInfo);   
  if(data2[1][1]!="") {
    tEnd = data2[1][1];  
    DebugTN("Overlapping: New tEnd", formatTime("%Y.%m.%d %H:%M:%S.", tEnd, "%03d"));
    ulLVSRecs++;
  } 

  return true;  
}

 



